/*
** ATOP - System & Process Monitor
**
** The program 'atop' offers the possibility to view the activity of 
** the system on system-level as well as process-level.
**
** This source-file contains the main-function, which verifies the
** calling-parameters and takes care of initialization. 
** The engine-function drives the main sample-loop in which after the
** indicated interval-time a snapshot is taken of the system-level and
** process-level counters and the deviations are calculated and
** visualized for the user.
** ================================================================
** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
** E-mail:      gerlof@ATComputing.nl
** Date:        November 1996
** Linux-port:  June 2000
** Modifications:
**		May 2001 - Ported to kernel 2.4
**
** After initialization, the main-function calls the ENGINE.
** For every cycle (so after another interval) the ENGINE calls various 
** functions as shown below:
**
** +---------------------------------------------------------------------+
** |                           E  N  G  I  N  E                          |
** |                                                                     |
** |                                                                     |
** |    _____________________await interval-timer_____________________   |
** |   |                                                              ^  |
** |   |      ________       ________      ________      ________     |  |
** |   |     ^        |     ^        |    ^        |    ^        |    |  |
** +---|-----|--------|-----|--------|----|--------|----|--------|----|--+
**     |     |        |     |        |    |        |    |        |    |
**  +--V-----|--+  +--V-----|--+  +--V----|--+  +--V----|--+  +--V----|-+  
**  |           |  |           |  |          |  |          |  |         |
**  | photosyst |  | photoproc |  |   acct   |  | deviate  |  |  print  |
**  |           |  |           |  |photoproc |  |  ...syst |  |         |
**  |           |  |           |  |          |  |  ...proc |  |         |
**  +-----------+  +-----------+  +----------+  +----------+  +---------+  
**        ^              ^             ^              ^            |
**        |              |             |              |            |
**        |              |             |              V            V 
**      ______       _________     __________     ________     _________
**     /      \     /         \   /          \   /        \   /         \
**      /proc          /proc       accounting     process-     screen or
**                                    file        database        file
**     \______/     \_________/   \__________/   \________/   \_________/
**
**    -	photosyst()
**	Takes a snapshot of the counters related to resource-usage on
** 	system-level (cpu, disk, memory, network).
**	This code is UNIX-flavor dependent; in case of Linux the counters
**	are retrieved from /proc.
**
**    -	photoproc()
**	Takes a snapshot of the counters related to resource-usage of
**	processes which are currently active. For this purpose the whole
**	task-list is read.
**	This code is UNIX-flavor dependent; in case of Linux the counters
**	are retrieved from /proc.
**
**    -	acctphotoproc()
**	Takes a snapshot of the counters related to resource-usage of
**	processes which have been finished during the last interval.
**	For this purpose all new records in the accounting-file are read.
**
** When all counters have been gathered, functions are called to calculate
** the difference between the current counter-values and the counter-values
** of the previous cycle. These functions operate on the system-level
** as well as on the process-level counters. 
** These differences are stored in a new structure(-table). 
**
**    -	deviatsyst()
**	Calculates the differences between the current system-level
** 	counters and the corresponding counters of the previous cycle.
**
**    -	deviatproc()
**	Calculates the differences between the current process-level
** 	counters and the corresponding counters of the previous cycle.
**	The per-process counters of the previous cycle are stored in the
**	process-database; this "database" is implemented as a linked list
**	of process-info structures in memory (so no disk-accesses needed).
**	Within this linked list hash-buckets are maintained for fast searches.
**	The entire process-database is handled via a set of well-defined 
** 	functions from which the name starts with "pdb_..." (see the
**	source-file procdbase.c).
**	The processes which have been finished during the last cycle
** 	are also treated by deviatproc() in order to calculate what their
**	resource-usage was before they finished.
**
** All information is ready to be visualized now.
** There is a structure which holds the start-address of the
** visualization-function to be called. Initially this structure contains
** the address of the generic visualization-function ("generic_samp"), but
** these addresses can be modified in the main-function depending on particular
** flags. In this way various representation-layers (ASCII, graphical, ...)
** can be linked with 'atop'; the one to use can eventually be chosen
** at runtime. 
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2, or (at your option) any
** later version.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** $Log: atop.c,v $
** Revision 1.8  2002/08/30 07:49:35  gerlof
** Implement possibility to store and retrieve atop-data in raw format.
**
** Revision 1.7  2002/08/27 12:09:16  gerlof
** Allow raw data file to be written and to be read (with compression).
**
** Revision 1.6  2002/07/24 11:12:07  gerlof
** Redesigned to ease porting to other UNIX-platforms.
**
** Revision 1.5  2002/07/11 09:15:53  root
** *** empty log message ***
**
** Revision 1.4  2002/07/08 09:20:45  root
** Bug solution: flag list overflow.
**
** Revision 1.3  2001/11/07 09:17:41  gerlof
** Use /proc instead of /dev/kmem for process-level statistics.
**
** Revision 1.2  2001/10/04 13:03:15  gerlof
** Separate kopen() function called i.s.o. implicit with first kmem-read
**
** Revision 1.1  2001/10/02 10:43:19  gerlof
** Initial revision
**
*/

static char rcsid[] = "$Id: atop.c,v 1.8 2002/08/30 07:49:35 gerlof Exp $";

#include <sys/types.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/utsname.h>
#include <string.h>
#include <zlib.h>

#include "atop.h"
#include "photoproc.h"
#include "photosyst.h"

#define	allflags	"ab:cde:fghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define	PROCCHUNK	100	/* portion-size for process-table expansion */
#define	MAXFL		64      /* maximum number of command-line flags     */
#define	BASEPATH	"/var/log/atop/"  
#define	MAXNAME		256

/*
** timing info
*/
time_t 		curtime;	/* might be used external by other functions */
time_t 		pretime;	/* might be used external by other functions */

static char	awaittrigger;	/* boolean: awaiting trigger for next cycle  */

/*
** argument values and global variables
*/
unsigned long	interval = 10;
unsigned int 	nsamples = 0xffffffff;
unsigned long	sampcnt;
int		deviatonly  = 1;

char		flaglist[MAXFL];
struct utsname	utsname;

/*
** structure which describes the raw file contents
**
** layout raw file:    rawheader
**
**                     rawrecord                           \
**                     compressed system-level statistics   | sample 1
**                     compressed process-level statistics /
**
**                     rawrecord                           \
**                     compressed system-level statistics   | sample 2
**                     compressed process-level statistics /
**
** etcetera .....
*/
#define	MYMAGIC	(unsigned long) 0xfeedbeef

struct rawheader {
	unsigned long	magic;
	unsigned short	sstatlen;
	unsigned short	pstatlen;
	unsigned short	rawreclen;
	char		acctfd;
	char		kernpatch;
	char		node[32];
	char		reserved[16];
};

struct rawrecord {
	time_t		curtime;	/* current time (epoch)         */
	unsigned short	interval;	/* interval (number of seconds) */
	unsigned short	nlist;		/* number of processes in list  */
	unsigned short	npresent;	/* total number of processes    */
	unsigned short	nexit;		/* number of exited processes   */
	unsigned short	nzombie;	/* number of zombie processes   */
	unsigned short	scomplen;	/* length of compressed sstat   */
	unsigned long	pcomplen;	/* length of compressed pstat's */
};

char		*rawname;
char		viewflag;
char		midnightflag;
unsigned int	begintime, endtime;

/*
** next struct contains pointers to four visualization-functions:
**   1.	show all info about one sample
**   2.	show error-message
**   3.	handle end-of-program
**   4.	show usage-info
*/
struct visualize vis = {generic_samp, generic_error, 
		        generic_end,  generic_usage};

/*
** internal prototypes
*/
static void	engine(void);
static void	rawread(void);
static void	rawwrite(time_t, int, struct sstat *, struct pstat *,
		         int, int, int, int);

int
main(int argc, char *argv[])
{
	register int	i;
	int		c;
	char		*p;

	/* 
	** interprete arguments & flags 
	*/
	if (argc > 1)
	{
		/* 
		** gather all flags for visualization-functions
		**
		** generic flags will be handled here;
		** unrecognized flags are passed to the print-routines
		*/
		i = 0;

		while (i < MAXFL-1 && (c=getopt(argc, argv, allflags)) != EOF)
		{
			switch (c)
			{
			   case '?':		/* usage wanted ?             */
				prusage(argv[0]);
				break;

			   case 'w':		/* writing of raw data ?      */
				if (optind >= argc)
					prusage(argv[0]);

				rawname       = argv[optind++];
				vis.show_samp = rawwrite;
				break;

			   case 'r':		/* reading of raw data ?      */
				if (optind < argc && *(argv[optind]) != '-')
					rawname = argv[optind++];

				viewflag++;
				break;

			   case 'S':		/* midnight limit ?           */
				midnightflag++;
				break;

                           case 'a':		/* all processes per sample ? */
				deviatonly=0;
				break;

                           case 'b':		/* begin time ?               */
				if ( !hhmm2secs(optarg, &begintime) )
					prusage(argv[0]);
				break;

                           case 'e':		/* end   time ?               */
				if ( !hhmm2secs(optarg, &endtime) )
					prusage(argv[0]);
				break;

			   default:		/* gather other flags */
				flaglist[i++] = c;
			}
		}

		/*
		** get optional interval-value and optional number of samples	
		*/
		if (optind < argc && optind < MAXFL)
		{
			if (!numeric(argv[optind]))
				prusage(argv[0]);
	
			interval = atoi(argv[optind]);
	
			optind++;
	
			if (optind < argc)
			{
				if (!numeric(argv[optind]) )
					prusage(argv[0]);

				if ( (nsamples = atoi(argv[optind])) < 1)
					prusage(argv[0]);
			}
		}
	}

	/*
	** check if raw data from a file must be viewed
	*/
	if (viewflag)
	{
		seteuid( getuid() );
		rawread();
		cleanstop(0);
	}

	/*
	** determine time since when current statistics have
	** been maintained
	*/
	curtime = getboot();

	/*
	** lock ATOP in memory to get reliable samples (also when
	** memory is low);
	** increment CPU scheduling-priority to get reliable samples (also
	** during heavy CPU load);
	** ignored if not running under superuser priviliges!
	*/
	(void) mlockall(MCL_FUTURE);
	(void) nice(-20);

	/*
	** catch signals for proper close-down
	*/
	signal(SIGHUP,  cleanstop);
	signal(SIGTERM, cleanstop);

	/*
	** switch-on the process-accounting mechanism to register the
	** (remaining) resource-usage by processes which have finished
	*/
	acctswon();

	/*
	** since process-accounting is switched on now, there is no
	** need to keep running under root-priviliges, so switch
	** effective user-id to real user-id
	*/
	seteuid ( getuid() );

	/*
	** determine the name of this node (without domain-name)
	*/
	(void) uname(&utsname);

	if ( (p = strchr(utsname.nodename, '.')) )
		*p = '\0';

	/*
	** start the engine now .....
	*/
	engine();

	cleanstop(0);
}

/*
** The engine() drives the main-loop of the program
*/
static void
engine(void)
{
	struct sigaction 	sigact;
	static time_t		timelimit;
	void			getalarm(int), getusr1(int);

	/*
	** reserve space for system-level statistics
	*/
	static struct sstat	*cursstat; /* current   */
	static struct sstat	*presstat; /* previous  */
	static struct sstat	*devsstat; /* deviation */
	static struct sstat	*hlpsstat;

	/*
	** reserve space for process-level statistics
	*/
	static struct pstat	*curpact;	/* current active list  */
	static int		curplen;	/* current active size  */

	struct pstat		*curpexit;	/* exitted process list	*/
	struct pstat		*devpstat;	/* deviation list	*/

	int			npresent, nexit, n, nzombie;

	/*
	** initialization: allocate required memory dynamically
	*/
	cursstat = calloc(1, sizeof(struct sstat));
	presstat = calloc(1, sizeof(struct sstat));
	devsstat = calloc(1, sizeof(struct sstat));

	/*
	** install the signal-handler for ALARM and SIGUSR1 (both triggers
	* for the next sample)
	*/
	memset(&sigact, 0, sizeof sigact);
	sigact.sa_handler = getusr1;
	sigaction(SIGUSR1, &sigact, (struct sigaction *)0);

	memset(&sigact, 0, sizeof sigact);
	sigact.sa_handler = getalarm;
	sigaction(SIGALRM, &sigact, (struct sigaction *)0);

	if (interval > 0)
		alarm(interval);

	if (midnightflag)
	{
		time_t		timenow = time(0);
		struct tm	*tp = localtime(&timenow);

		tp->tm_hour = 23;
		tp->tm_min  = 59;
		tp->tm_sec  = 59;

		timelimit = mktime(tp);
	}

	/*
	** MAIN-LOOP:
	**    -	Wait for the requested number of seconds or for other trigger
	**
	**    -	System-level counters
	**		get current counters
	**		calculate the differences with the previous sample
	**
	**    -	Process-level counters
	**		get current counters from running & exited processes
	**		calculate the differences with the previous sample
	**
	**    -	Call the print-function to visualize the differences
	*/
	for (sampcnt=0; sampcnt < nsamples; sampcnt++)
	{
		/*
		** if the limit-flag is specified:
		**  check if the next sample is expected before midnight;
		**  if not, stop atop now 
		*/
		if (midnightflag && (curtime+interval) > timelimit)
			break;

		/*
		** wait for alarm-signal to arrive (except first sample)
		** or wait for SIGUSR1 in case of an interval of 0.
		*/
		if (sampcnt > 0 && awaittrigger)
			pause();

		awaittrigger = 1;

		/*
		** gather time info for this sample
		*/
		pretime  = curtime;
		curtime  = time(0);		/* seconds since 1-1-1970 */

		/*
		** take a snapshot of the current system-level statistics 
		** and calculate the deviations (i.e. calculate the activity
		** during the last sample)
		*/
		hlpsstat = cursstat;	/* swap current/prev. stats */
		cursstat = presstat;
		presstat = hlpsstat;

		photosyst(cursstat);	/* obtain new counters      */

		deviatsyst(cursstat, presstat, devsstat);

		/*
		** take a snapshot of the current process-level statistics 
		** and calculate the deviations (i.e. calculate the activity
		** during the last sample)
		**
		** first register active processes
		**  --> atop malloc's a minimal amount of space which is
		**      only extended when needed
		*/
		while ( (npresent = photoproc(curpact, curplen)) == curplen)
		{
			curplen += PROCCHUNK;

			curpact = realloc(curpact,
			                   curplen * sizeof(struct pstat));
			memset(curpact, 0, curplen * sizeof(struct pstat));
		}

		/*
		** register processes which exited during last sample;
		** first determine how many processes exited and
		** reserve space for them, and secondly obtain the info
		*/
		nexit = acctprocnt();	/* number of exited processes */

		if (nexit > 0)	
		{
			curpexit = malloc(nexit * sizeof(struct pstat));
			acctphotoproc(curpexit, nexit);
		}

		/*
		** calculate deviations
		*/
		devpstat = malloc((npresent+nexit) * sizeof(struct pstat));

		n = deviatproc(curpact, npresent, curpexit, nexit,
					deviatonly, devpstat, &nzombie);

		/*
		** activate the installed print-function to visualize
		** the deviations
		*/
		(vis.show_samp)( curtime,
				 curtime-pretime > 0 ? curtime-pretime : 1,
		           	 devsstat, devpstat, n, npresent,
		           	 nzombie, nexit);

		/*
		** release dynamically allocated memory
		*/
		if (nexit > 0)
			free(curpexit);

		free(devpstat);
	} /* end of main-loop */
}

/*
** print usage of this command
*/
void
prusage(char *myname)
{
	printf("Usage: %s [-flags] [interval [samples]]\n", myname);
	printf("\t\tor\n");
	printf("Usage: %s -w  file  [-S] [-a] [interval [samples]]\n", myname);
	printf("       %s -r [file] [-b hh:mm] [-e hh:mm] [-flags]\n", myname);
	printf("\n");
	printf("\tgeneric flags:\n");
	printf("\t  -a  show or log all processes (i.s.o. active processes "
	                "only)\n");

	(*vis.show_usage)();

	printf("\n");
	printf("\tspecific flags for raw logfiles:\n");
	printf("\t  -w  write raw data to   file (compressed)\n");
	printf("\t  -r  read  raw data from file (compressed)\n");
	printf("\t  -S  finish atop automatically before midnight "
	                "(i.s.o. #samples)\n");
	printf("\t  -b  begin showing data from specified time\n");
	printf("\t  -e  finish showing data after specified time\n");
	printf("\n");
	printf("\tinterval: number of seconds   (minimum 0)\n");
	printf("\tsamples:  number of intervals (minimum 1)\n");
	printf("\n");
	printf("If the interval-value is zero, a new sample can be\n");
	printf("forced manually by sending signal USR1"
			" (kill -USR1 pid_atop)\n");
	printf("or with the keystroke 't' in interactive mode.\n");


	cleanstop(1);
}

/*
** handler for ALRM-signal
*/
void
getalarm(int sig)
{
	awaittrigger=0;

	if (interval > 0)
		alarm(interval);	/* restart the timer */
}

/*
** handler for USR1-signal
*/
void
getusr1(int sig)
{
	awaittrigger=0;
}

/*
** write a raw record to file
** (file is created during the first call)
*/
static void
rawwrite(time_t curtime, int numsecs, 
	 struct sstat *ss, struct pstat *ps,
	 int nlist, int npresent, int nzombie, int nexit)
{
	static int		rawfd = -1;
	struct rawrecord	rr;

	Byte			scompbuf[sizeof(struct sstat)], *pcompbuf;
	unsigned long		scomplen = sizeof scompbuf;
	unsigned long		pcomplen = sizeof(struct pstat) * nlist;

	/*
	** first call:
	**   - create the raw file
	**   - write header record
	*/
	if (rawfd == -1)
	{
		struct rawheader	rh;
		extern int		acctfd;
		extern char		kernpatch;
		time_t 			timenow;

		if ( (rawfd = creat(rawname, 0666)) == -1)
		{
			perror("create raw file");
			cleanstop(7);
		}

		rh.magic	= MYMAGIC;
		rh.sstatlen	= sizeof(struct sstat);
		rh.pstatlen	= sizeof(struct pstat);
		rh.rawreclen	= sizeof(struct rawrecord);
		rh.acctfd	= acctfd;
		rh.kernpatch	= kernpatch;

		strncpy(rh.node, utsname.nodename, sizeof rh.node);
		rh.node[sizeof rh.node -1] = '\0';

		memset(rh.reserved, 0, sizeof rh.reserved);

		if ( write(rawfd, &rh, sizeof rh) == -1)
		{
			perror("write raw header");
			cleanstop(7);
		}
	}

	/*
	** compress system- and process-level statistics
	*/
	(void) compress(scompbuf, &scomplen,
				(Byte *)ss, (unsigned long)sizeof *ss);

	pcompbuf = malloc(pcomplen);

	(void) compress(pcompbuf, &pcomplen,
				(Byte *)ps, (unsigned long)pcomplen);

	/*
	** fill record header and write to file
	*/
	rr.curtime	= curtime;
	rr.interval	= numsecs;
	rr.nlist	= nlist;
	rr.npresent	= npresent;
	rr.nzombie	= nzombie;
	rr.nexit	= nexit;
	rr.scomplen	= scomplen;
	rr.pcomplen	= pcomplen;

	if ( write(rawfd, &rr, sizeof rr) == -1)
	{
		perror("write raw record");
		cleanstop(7);
	}

	/*
	** write compressed system status structure to file
	*/
	if ( write(rawfd, scompbuf, scomplen) == -1)
	{
		perror("write raw status record");
		cleanstop(7);
	}

	/*
	** write compressed list of process status structures to file
	*/
	if ( write(rawfd, pcompbuf, pcomplen) == -1)
	{
		perror("write raw process record");
		cleanstop(7);
	}

	free(pcompbuf);
}

/*
** read the contents of a raw file
*/
static void
rawread(void)
{
	int			rawfd;
	extern char		kernpatch;
	extern int		acctfd;
	struct rawheader	rh;
	struct rawrecord	rr;
	struct sstat		devsstat;
	struct pstat		*devpstat;

	Byte			*compbuf;
	unsigned long		uncomplen;

	/*
	** if no filename is specified, assemble the name of the raw file
	*/
	if (!rawname)
	{
		time_t		timenow;
		struct tm	*tp;

		timenow	= time(0);
		tp	= localtime(&timenow);

		rawname = malloc(MAXNAME);

		snprintf(rawname, MAXNAME, "%s/atop_%04d%02d%02d",
			BASEPATH, 
			tp->tm_year+1900,
			tp->tm_mon+1,
			tp->tm_mday);
	}

	/*
	** open raw file
	*/
	if ( (rawfd = open(rawname, O_RDONLY)) == -1)
	{
		char	command[512], tmpname[64];

		/*
		** check if a compressed raw file is present
		*/
		strncat(rawname, ".gz", MAXNAME - strlen(rawname));

		if ( access(rawname, F_OK|R_OK) == -1)
		{
			perror("access raw file");
			cleanstop(7);
		}

		/*
		** compressed raw file to be decompressed via gunzip
		*/
		fprintf(stderr, "Decompressing logfile ....\n");

		snprintf(tmpname, sizeof tmpname, "/tmp/atopwrk_%d", getpid());
		snprintf(command, sizeof command, "gunzip -c %s > %s",
							rawname, tmpname);
		system (command);

		if ( (rawfd = open(tmpname, O_RDONLY)) == -1)
		{
			perror("open decompressed raw file");
			cleanstop(7);
		}

		unlink(tmpname);
	}

	/*
	** read the raw header and verify the magic number and version
	*/
	if ( read(rawfd, &rh, sizeof rh) < sizeof rh)
	{
		fprintf(stderr, "can not read raw file header\n");
		cleanstop(7);
	}

	if (rh.magic != MYMAGIC)
	{
		fprintf(stderr, "no raw atop output (wrong magic number)\n");
		cleanstop(7);
	}

	if (  rh.sstatlen  != sizeof(struct sstat)	||
	      rh.pstatlen  != sizeof(struct pstat)	||
	      rh.rawreclen != sizeof(struct rawrecord)	  )
	{
		fprintf(stderr,
			"raw file %s is created by incompatible version "
		        "of atop\n", rawname);
		cleanstop(7);
	}

	strncpy(utsname.nodename, rh.node, sizeof utsname.nodename);
	utsname.nodename[sizeof utsname.nodename -1] = '\0';

	kernpatch = rh.kernpatch;
	acctfd    = rh.acctfd;
	interval  = 0;

	/*
	** read a raw record header until end-of-file
	*/
	while ( read(rawfd, &rr, sizeof rr) == sizeof rr)
	{
		unsigned int	secsinday = daysecs(rr.curtime);

		/*
		** check if this sample is within the time-range
		** specified with the -b and -e flags (if any)
		*/
		if ( (begintime && begintime > secsinday) ||
		     (endtime   && endtime   < secsinday)   )
		{
			lseek(rawfd, rr.scomplen+rr.pcomplen, SEEK_CUR);
			continue;
		}

		/*
		** allocate space, read compressed system-level statistics
		** and decompress
		*/
		compbuf = malloc(rr.scomplen);

		if ( read(rawfd, compbuf, rr.scomplen) < rr.scomplen)
			break;

		uncomplen = sizeof devsstat;

		(void) uncompress((Byte *)&devsstat, &uncomplen,
				  		compbuf, rr.scomplen);

		free(compbuf);

		/*
		** allocate space, read compressed process-level statistics
		** and decompress
		*/
		compbuf = malloc(rr.pcomplen);

		if ( read(rawfd, compbuf, rr.pcomplen) < rr.pcomplen)
			break;

		uncomplen = sizeof(struct pstat) * rr.nlist;
		devpstat  = malloc(uncomplen);

		(void) uncompress((Byte *)devpstat, &uncomplen,
				  		compbuf, rr.pcomplen);

		free(compbuf);

		/*
		** activate the installed print-function to visualize
		** the system- and process-level statistics
		*/
		(vis.show_samp)( rr.curtime, rr.interval, &devsstat, devpstat,
				 rr.nlist, rr.npresent, rr.nzombie, rr.nexit);

		free(devpstat);
	}

	close(rawfd);
	cleanstop(0);
}
