/*
** ATOP - System & Process Monitor
**
** The program 'atop' offers the possibility to view the activity of
** the system on system-level as well as process-level.
**
** This source-file contains functions to calculate the differences for
** the system-level and process-level counters since the previous sample.
** ================================================================
** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
** E-mail:      gerlof@ATComputing.nl
** Date:        November 1996
** LINUX-port:  June 2000
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2, or (at your option) any
** later version.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** $Log: deviate.c,v $
** Revision 1.8  2002/08/27 04:47:46  gerlof
** Minor comment updates.
**
** Revision 1.7  2002/07/24 11:12:20  gerlof
** Redesigned to ease porting to other UNIX-platforms.
**
** Revision 1.6  2002/07/10 04:59:37  root
** Counters pin/pout renamed to swin/swout (Linux conventions).
**
** Revision 1.5  2002/01/22 13:39:20  gerlof
** Support for number of cpu's.
**
** Revision 1.4  2001/11/22 08:33:10  gerlof
** Add priority per process.
**
** Revision 1.3  2001/11/07 09:18:22  gerlof
** Use /proc instead of /dev/kmem for process-level statistics.
**
** Revision 1.2  2001/10/03 08:58:41  gerlof
** Improved subtraction which is overflow-proof
**
** Revision 1.1  2001/10/02 10:43:23  gerlof
** Initial revision
**
*/

static char rcsid[] = "$Id: deviate.c,v 1.8 2002/08/27 04:47:46 gerlof Exp $";

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>

#include "atop.h"
#include "photoproc.h"
#include "photosyst.h"

extern time_t	pretime;	/* previous sample time 	 */

/*
** calculate the process-activity during the last sample
*/
int
deviatproc(struct pstat *aproc, int npresent,
           struct pstat *eproc, int nexit, int deviatonly,
	   struct pstat *dproc, int *nzombie)
{
	register int		c, d, i;
	register struct pstat	*curstat, *devstat;
	struct pstat		prestat;
	struct pinfo		*pinfo;

	/*
	** make new list of all processes in the process-database;
	** after handling all processes, the left-overs
	** should have disappeared since the previous sample
	*/
	pdb_makeresidue();

	/*
	** calculate deviations per present process
	*/
	for (c=0, d=0, *nzombie=0; c < npresent; c++)
	{
		curstat = aproc+c;

		if (curstat->gen.state == 'Z')
			(*nzombie)++;

		/*
		** get previous figures from process-database
		*/
		if ( pdb_getproc(curstat->gen.pid, &pinfo) )
		{
			/*
			** process already present during the previous sample;
			** check differences with previous sample
			*/
			if ( deviatonly && memcmp(curstat, &pinfo->pstat, 
					           sizeof(struct pstat)) == EQ)
				continue;

			/*
			** differences detected, so the process was active or
			** its status or memory-occupation has changed;
			** save stats from previous sample (to use for
			** further calculations) and store new statistics
			** in process-database
			*/
			prestat 	= pinfo->pstat;	/* save old	*/
			pinfo->pstat 	= *curstat;	/* overwrite	*/
		}
		else
		{
			/*
			** new process which must have been started during
			** last interval
			*/
			memset(&prestat, 0, sizeof(prestat));

			/*
			** create new process
			*/
			pdb_newproc(&pinfo);

			pinfo->pstat = *curstat;

			/*
			** add new process to process-database
			*/
			pdb_addproc( curstat->gen.pid, pinfo);
		}

		/*
		** now do the calculations
		*/
		devstat = dproc+d;

		devstat->gen        = curstat->gen;
		devstat->cpu.pri    = curstat->cpu.pri;

		devstat->cpu.stime  = 
			subtrul(curstat->cpu.stime, prestat.cpu.stime);
		devstat->cpu.utime  =
			subtrul(curstat->cpu.utime, prestat.cpu.utime);

		devstat->dsk.ior    =
			subtrul(curstat->dsk.ior, prestat.dsk.ior);
		devstat->dsk.iow    =
			subtrul(curstat->dsk.iow, prestat.dsk.iow);

		devstat->mem.shtext = curstat->mem.shtext;
		devstat->mem.vmem   = curstat->mem.vmem;
		devstat->mem.rmem   = curstat->mem.rmem;
		devstat->mem.vgrow  = curstat->mem.vmem   - prestat.mem.vmem;
		devstat->mem.rgrow  = curstat->mem.rmem   - prestat.mem.rmem;
		devstat->mem.minflt = 
			subtrul(curstat->mem.minflt, prestat.mem.minflt);
		devstat->mem.majflt =
			subtrul(curstat->mem.majflt, prestat.mem.majflt);
		devstat->mem.iosw   =
			subtrul(curstat->mem.iosw, prestat.mem.iosw);

		devstat->net.tcpsnd =
			subtrul(curstat->net.tcpsnd, prestat.net.tcpsnd);
		devstat->net.tcprcv =
			subtrul(curstat->net.tcprcv, prestat.net.tcprcv);
		devstat->net.udpsnd =
			subtrul(curstat->net.udpsnd, prestat.net.udpsnd);
		devstat->net.udprcv =
			subtrul(curstat->net.udprcv, prestat.net.udprcv);
		d++;
	}

	/*
	** calculate deviations per exited process
	*/
	for (c=0; c < nexit; c++)
	{
		/*
		** check if this process has been started AND
		** finished since previous sample;
		** if so, it has no use to check if there is still 
		** existing info present in the process-database
		*/
		curstat = eproc+c;

		if ( curstat->gen.btime > pretime )
		{
			/*
			** process-start and -finish during same interval
			*/
			memset(&prestat, 0, sizeof(prestat));
		}
		else
		{
			/*
			** process must be known in process-database;
			** try to match one of the remaining processes
			** against this exited one
			*/
			if ( pdb_srchresidue(curstat, &pinfo) )
				prestat = pinfo->pstat;
			else
				memset(&prestat, 0, sizeof(prestat));
		}

		/*
		** now do the calculations
		*/
		devstat = dproc+d;

		devstat->gen        = curstat->gen;
		devstat->gen.pid    = prestat.gen.pid;
		devstat->cpu.pri    = 0;

		devstat->cpu.stime  =
			subtrul(curstat->cpu.stime, prestat.cpu.stime);
		devstat->cpu.utime  =
			subtrul(curstat->cpu.utime, prestat.cpu.utime);

		devstat->dsk.iow    = 0;
		devstat->dsk.ior    = curstat->dsk.ior    - prestat.dsk.ior -
							    prestat.dsk.iow;

		devstat->mem.iosw   =
			subtrul(curstat->mem.iosw, prestat.mem.iosw);
		devstat->mem.shtext = 0;
		devstat->mem.vmem   = 0;
		devstat->mem.rmem   = 0;
		devstat->mem.vgrow  = 0;
		devstat->mem.rgrow  = 0;
		devstat->mem.minflt =
			subtrul(curstat->mem.minflt, prestat.mem.minflt);
		devstat->mem.majflt =
			subtrul(curstat->mem.majflt, prestat.mem.majflt);

		devstat->net.tcpsnd = 0;
		devstat->net.tcprcv = 0;
		devstat->net.udpsnd = 0;
		devstat->net.udprcv = 0;

		if (devstat->cpu.stime < 0)
			devstat->cpu.stime = 0;
		if (devstat->cpu.utime < 0)
			devstat->cpu.utime = 0;

		d++;

		if (prestat.gen.pid > 0)
			pdb_delproc(prestat.gen.pid);
	}

	/*
	** remove unused entries from RESIDUE chain
	*/
	pdb_cleanresidue();

	return d;
}

/*
** calculate the system-activity during the last sample
*/
void
deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
{
	register int	i;

	dev->cpu.stime 		= subtrul(cur->cpu.stime,  pre->cpu.stime);
	dev->cpu.utime 		= subtrul(cur->cpu.utime,  pre->cpu.utime);
	dev->cpu.ntime		= subtrul(cur->cpu.ntime,  pre->cpu.ntime);
	dev->cpu.itime 		= subtrul(cur->cpu.itime,  pre->cpu.itime);
	dev->cpu.devint		= subtrul(cur->cpu.devint, pre->cpu.devint);
	dev->cpu.csw		= subtrul(cur->cpu.csw,    pre->cpu.csw);
	dev->cpu.numcpu		= cur->cpu.numcpu;

	dev->mem.physmem	= cur->mem.physmem;
	dev->mem.freemem	= cur->mem.freemem;
	dev->mem.buffermem	= cur->mem.buffermem;
	dev->mem.sharedmem	= cur->mem.sharedmem;
	dev->mem.cachemem	= cur->mem.cachemem;
	dev->mem.totswap	= cur->mem.totswap;
	dev->mem.freeswap	= cur->mem.freeswap;

	dev->mem.swouts		= subtrul(cur->mem.swouts,  pre->mem.swouts);
	dev->mem.swins		= subtrul(cur->mem.swins,   pre->mem.swins);

	dev->net.ipin 		= subtrul(cur->net.ipin,    pre->net.ipin);
	dev->net.ipout 		= subtrul(cur->net.ipout,   pre->net.ipout);
	dev->net.ipindel 	= subtrul(cur->net.ipindel, pre->net.ipindel);
	dev->net.ipfrw 		= subtrul(cur->net.ipfrw,   pre->net.ipfrw);
	dev->net.tcpin 		= subtrul(cur->net.tcpin,   pre->net.tcpin);
	dev->net.tcpout		= subtrul(cur->net.tcpout,  pre->net.tcpout);
	dev->net.udpin 		= subtrul(cur->net.udpin,   pre->net.udpin);
	dev->net.udpout		= subtrul(cur->net.udpout,  pre->net.udpout);

	if (pre->intf[0].name[0] == '\0')	/* first sample? */
	{
		for (i=0; cur->intf[i].name[0]; i++)
			strcpy(pre->intf[i].name, cur->intf[i].name);
	}
	
	for (i=0; cur->intf[i].name[0]; i++)
	{
		/*
		** check if an interface has been added or removed;
		** in that case, skip further handling for this sample
		*/
		if (strcmp(cur->intf[i].name, pre->intf[i].name) != 0)
			break;

		strcpy(dev->intf[i].name, cur->intf[i].name);

		dev->intf[i].rbyte  =
				subtrul(cur->intf[i].rbyte, pre->intf[i].rbyte);
		dev->intf[i].rpack  =
				subtrul(cur->intf[i].rpack, pre->intf[i].rpack);
		dev->intf[i].sbyte  =
				subtrul(cur->intf[i].sbyte, pre->intf[i].sbyte);
		dev->intf[i].spack  =
				subtrul(cur->intf[i].spack, pre->intf[i].spack);
	}
	dev->intf[i].name[0] = '\0';

	for (i=0; cur->xdsk[i].name[0]; i++)
	{
		strcpy(dev->xdsk[i].name, cur->xdsk[i].name);
		dev->xdsk[i].nread  =
			subtrul(cur->xdsk[i].nread,  pre->xdsk[i].nread);
		dev->xdsk[i].nwrite =
			subtrul(cur->xdsk[i].nwrite, pre->xdsk[i].nwrite);
		dev->xdsk[i].nrblk  =
			subtrul(cur->xdsk[i].nrblk,  pre->xdsk[i].nrblk);
		dev->xdsk[i].nwblk  =
			subtrul(cur->xdsk[i].nwblk,  pre->xdsk[i].nwblk);
		dev->xdsk[i].io_ms  =
			subtrul(cur->xdsk[i].io_ms,  pre->xdsk[i].io_ms);
		dev->xdsk[i].avque  =
			subtrul(cur->xdsk[i].avque,  pre->xdsk[i].avque);
	}
	dev->xdsk[i].name[0] = '\0';
}

/*
** Generic function to subtract two unsigned long's taking into 
** account the possibility of overflow.
*/
unsigned long
subtrul(unsigned long newval, unsigned long oldval)
{
	static double	ulongmax = ULONG_MAX + 1.0;

	if (newval >= oldval)
		return newval - oldval;
	else
		return (unsigned long) (ulongmax + newval - oldval);
}
